@@ -16,6 +16,8 @@
 from sageopt import coniclifts
 from sageopt import relaxations
 from sageopt import symbolic
+from sageopt.symbolic import signomials
+from sageopt.symbolic import polynomials
 
 from sageopt.relaxations import local_refine, local_refine_polys_from_sigs
 from sageopt.relaxations import infer_domain
@@ -29,4 +31,7 @@ from sageopt.symbolic.polynomials import Polynomial, standard_poly_monomials, Po
 from sageopt.relaxations import poly_relaxation, poly_constrained_relaxation
 from sageopt.relaxations import poly_solrec
 
-__version__ = '0.4.1'
+import sageopt.interop.scipy
+import sageopt.interop.cvxpy
+
+__version__ = '0.5.1'
