@@ -19,24 +19,32 @@ from sageopt.coniclifts.constraints.set_membership.product_cone import PrimalPro
 from sageopt.coniclifts.base import Variable, Expression
 from sageopt.coniclifts.cones import Cone
 from sageopt.coniclifts.operators import affine as aff
+from sageopt.coniclifts.operators.pos import pos as pos_operator
 from sageopt.coniclifts.operators.norms import vector2norm
 from sageopt.coniclifts.operators.precompiled.relent import sum_relent, elementwise_relent
 from sageopt.coniclifts.operators.precompiled import affine as comp_aff
 from sageopt.coniclifts.problems.problem import Problem
-from sageopt.coniclifts.standards.constants import minimize as CL_MIN, solved as CL_SOLVED
+from sageopt.coniclifts.standards.constants import (
+    minimize as CL_MIN,
+    solved as CL_SOLVED,
+    inaccurate as CL_INACCURATE
+)
+from sageopt.coniclifts.utilities import kernel_basis
 import warnings
 import scipy.special as special_functions
 
 
 _ALLOWED_CONES_ = {'+', 'S', 'e', '0'}
 
-_AGGRESSIVE_REDUCTION_ = True
 
-_ELIMINATE_TRIVIAL_AGE_CONES_ = False
-
-_REDUCTION_SOLVER_ = 'ECOS'
-
-_SUM_AGE_FORCE_EQUALITY_ = False
+SETTINGS = {
+    'heuristic_reduction': True,
+    'presolve_trivial_age_cones': False,
+    'reduction_solver': 'ECOS',
+    'sum_age_force_equality': False,
+    'compact_dual': False,
+    'kernel_basis': False,
+}
 
 
 def check_cones(K):
@@ -47,7 +55,18 @@ def check_cones(K):
 
 class PrimalSageCone(SetMembership):
     """
-    Represent the constraint ":math:`c \\in C_{\\mathrm{SAGE}}(\\alpha, X)`".
+    Require that :math:`f(x) = \\sum_{i=1}^m c_i \\exp(\\alpha_i \\cdot x)` is nonnegative on
+    :math:`X`, and furthermore that we can *prove* this nonnegativity by decomposing :math:`f`
+    into a sum of ":math:`X`-AGE functions." It is useful to summarize this constraint by writing
+
+    .. math::
+
+        c \\in C_{\\mathrm{SAGE}}(\\alpha, X).
+
+    Each :math:`X`-AGE function used in the decomposition of :math:`f` can be proven nonnegative
+    by certifying a particular relative entropy inequality. This PrimalSageCone object tracks
+    both coefficients of :math:`X`-AGE functions used in the decomposition of :math:`f`,
+    and the additional variables needed to certify the associated relative entropy inequalities.
 
     Parameters
     ----------
@@ -77,6 +96,7 @@ class PrimalSageCone(SetMembership):
 
         ``covers[i]`` indicates which indices ``j`` have ``alpha[j,:]`` participate in
         the i-th AGE cone. Automatically constructed in a presolve phase, if not provided.
+        See also ``PrimalSageCone.ech``.
 
 
     Attributes
@@ -92,24 +112,54 @@ class PrimalSageCone(SetMembership):
 
     age_vectors : Dict[int, Expression]
 
-        If all Variable objects in the scope of this constraint are assigned feasible,
-        values, then we should have ``age_vectors[i].value`` in the i-th AGE cone with
-        respect to ``alpha``, and ``c.value == sum([av.value for av in age_vectors.values()], axis=0)``.
+        A vector :math:`c^{(i)} = \\mathtt{age{\\_}vectors[i]}` can have at most one negative
+        component :math:`c^{(i)}_i`. Such a vector defines a signomial
+
+        .. math::
+
+          f_i(x) = \\textstyle\\sum_{j=1}^m c^{(i)}_j \\exp(\\alpha_j \\cdot x)
+
+        which is nonnegative for all :math:`x \\in X`.
+        Letting :math:`N \\doteq \\mathtt{self.age\\_vectors.keys()}`, we should have
+
+        .. math::
+
+          \\mathtt{self.c} \\geq \\textstyle\\sum_{i \\in N} \\mathtt{self.age\\_vectors[i]}.
+
+        See also ``age_witnesses``.
+
+    age_witnesses : Dict[int, Expression]
+
+        A vector :math:`w^{(i)} = \\mathtt{age{\\_}witnesses[i]}` should certify that the signomial
+        with coefficients :math:`c^{(i)} = \\mathtt{age{\\_}vectors[i]}` is nonnegative on :math:`X`.
+        Specifically, we should have
+
+        .. math::
+
+          \\sigma_X\\left(-\\alpha^\\intercal w^{(i)}\\right)
+            + D\\left(w^{(i)}_{\\setminus i}, e c^{(i)}_{\\setminus i}\\right)\\leq c^{(i)}_i
+
+        where :math:`\\sigma_X` is the support function of :math:`X`, :math:`e` is Euler's number,
+        :math:`w^{(i)}_{\\setminus i} = ( w^{(i)}_j )_{j \\in [m]\\setminus \\{i\\}}`,
+        :math:`c^{(i)}_{\\setminus i} = ( c^{(i)}_j )_{j \\in [m]\\setminus \\{i\\}}`,
+        and :math:`D(\\cdot,\\cdot)` is the relative entropy function.
+        The validity of this certificate stems from the weak-duality principle in convex optimization.
 
     X : SigDomain or None
 
         If None, then this constraint represents a primal ordinary-SAGE cone.
-
-    ech : ExpCoverHelper
-
-        A simple wrapper around the constructor argument ``covers``. Manages validation of ``covers``
-        when provided, and manages construction of ``covers`` when a user does not provide it.
+        See also the function ``PrimalSageCone.sigma_x``.
 
     name : str
 
         Uniquely identifies this Constraint in the model where it appears. Serves as a suffix
         for the name of any auxiliary Variable created when compiling to the coniclifts standard.
 
+    ech : ExpCoverHelper
+
+        A simple wrapper around the constructor argument ``covers``. Validates ``covers``
+        when provided, and constructs ``covers`` when a user does not provide it.
+
     Notes
     -----
 
@@ -118,7 +168,10 @@ class PrimalSageCone(SetMembership):
     """
 
     def __init__(self, c, alpha, X, name, **kwargs):
+        self.settings = SETTINGS.copy()
         covers = kwargs['covers'] if 'covers' in kwargs else None
+        if 'settings' in kwargs:
+            self.settings.update(kwargs['settings'])
         self._n = alpha.shape[1]
         self._m = alpha.shape[0]
         self.name = name
@@ -128,50 +181,90 @@ class PrimalSageCone(SetMembership):
         if X is not None:
             check_cones(X.K)
             self._lifted_n = X.A.shape[1]
-            self.ech = ExpCoverHelper(self.alpha, self.c, (X.A, X.b, X.K), covers)
+            self.ech = ExpCoverHelper(self.alpha, self.c, (X.A, X.b, X.K), covers, self.settings)
         else:
             self._lifted_n = self._n
-            self.ech = ExpCoverHelper(self.alpha, self.c, None, covers)
+            self.ech = ExpCoverHelper(self.alpha, self.c, None, covers, self.settings)
         self.age_vectors = dict()
-        self._nu_vars = dict()
+        self._sfx = None  # "suppfunc x"; for evaluating support function
+        self._age_witnesses = None
+        self._nus = dict()
+        self._pre_nus = dict()  # if kernel_basis == True, then store certain Variables here.
+        self._nu_bases = dict()
         self._c_vars = dict()
         self._relent_epi_vars = dict()
         self._eta_vars = dict()
-        self._initialize_variables()
+        self._variables = self.c.variables()
+        if self._m > 1 and self.X is None:
+            self._ordsage_init_variables()
+        elif self._m > 1:
+            self._condsage_init_variables()
         self._build_aligned_age_vectors()
         pass
 
-    def _initialize_variables(self):
-        self._variables = self.c.variables()
-        if self._m > 1:
-            for i in self.ech.U_I:
-                num_cover = self.ech.expcover_counts[i]
-                if num_cover > 0:
-                    var_name = 'nu^{(' + str(i) + ')}_' + self.name
-                    self._nu_vars[i] = Variable(shape=(num_cover,), name=var_name)
-                    var_name = '_relent_epi_^{(' + str(i) + ')}_' + self.name
-                    self._relent_epi_vars[i] = Variable(shape=(num_cover,), name=var_name)
-                    if self.X is not None:
-                        var_name = 'eta^{(' + str(i) + ')}_{' + self.name + '}'
-                        self._eta_vars[i] = Variable(shape=(self.X.b.size,), name=var_name)
-                c_len = num_cover
-                if i not in self.ech.N_I:
-                    c_len += 1
-                if c_len == 0:  # pragma: no cover
-                    msg = 'PrimalSageCone constraint "' + self.name + '" encountered an index '
-                    msg += 'i=' + str(i) + '\n where the i-th AGE cone reduces to the nonnegative '
-                    msg += 'orthant, but self.c[i]=' + str(self.c[i].value) + ' is negative. \n\n'
-                    msg += 'This SAGE constraint is infeasible.'
-                    raise RuntimeError(msg)
-                var_name = 'c^{(' + str(i) + ')}_{' + self.name + '}'
-                self._c_vars[i] = Variable(shape=(c_len,), name=var_name)
-            self._variables += list(self._nu_vars.values())
-            self._variables += list(self._c_vars.values())
-            self._variables += list(self._relent_epi_vars.values())
-            if self.X is not None:
-                self._variables += list(self._eta_vars.values())
+    def _condsage_init_variables(self):
+        for i in self.ech.U_I:
+            num_cover = self.ech.expcover_counts[i]
+            if num_cover > 0:
+                var_name = 'nu^{(' + str(i) + ')}_' + self.name
+                self._nus[i] = Variable(shape=(num_cover,), name=var_name)
+                var_name = '_relent_epi_^{(' + str(i) + ')}_' + self.name
+                self._relent_epi_vars[i] = Variable(shape=(num_cover,), name=var_name)
+                var_name = 'eta^{(' + str(i) + ')}_{' + self.name + '}'
+                self._eta_vars[i] = Variable(shape=(self.X.b.size,), name=var_name)
+            c_len = self._c_len_check_infeas(num_cover, i)
+            var_name = 'c^{(' + str(i) + ')}_{' + self.name + '}'
+            self._c_vars[i] = Variable(shape=(c_len,), name=var_name)
+        self._variables += list(self._nus.values())
+        self._variables += list(self._c_vars.values())
+        self._variables += list(self._relent_epi_vars.values())
+        self._variables += list(self._eta_vars.values())
         pass
 
+    def _ordsage_init_variables(self):
+        for i in self.ech.U_I:
+            num_cover = self.ech.expcover_counts[i]
+            if num_cover > 0:
+                var_name = 'nu^{(' + str(i) + ')}_' + self.name
+                if self.settings['kernel_basis']:
+                    var_name = '_pre_' + var_name
+                    idx_set = self.ech.expcovers[i]
+                    mat = (self.alpha[idx_set, :] - self.alpha[i, :]).T
+                    nu_i_basis = kernel_basis(mat)
+                    self._nu_bases[i] = nu_i_basis
+                    pre_nu_i = Variable(shape=(nu_i_basis.shape[1],), name=var_name)
+                    self._pre_nus[i] = pre_nu_i
+                    self._nus[i] = self._nu_bases[i] @ pre_nu_i
+                else:
+                    var_name = 'nu^{(' + str(i) + ')}_' + self.name
+                    nu_i = Variable(shape=(num_cover,), name=var_name)
+                    self._nus[i] = nu_i
+                var_name = '_relent_epi_^{(' + str(i) + ')}_' + self.name
+                self._relent_epi_vars[i] = Variable(shape=(num_cover,), name=var_name)
+            c_len = self._c_len_check_infeas(num_cover, i)
+            var_name = 'c^{(' + str(i) + ')}_{' + self.name + '}'
+            self._c_vars[i] = Variable(shape=(c_len,), name=var_name)
+        if self.settings['kernel_basis']:
+            self._variables += list(self._pre_nus.values())
+        else:
+            self._variables += list(self._nus.values())
+        self._variables += list(self._c_vars.values())
+        self._variables += list(self._relent_epi_vars.values())
+        pass
+
+    # pragma: no cover
+    def _c_len_check_infeas(self, num_cover, i):
+        c_len = num_cover
+        if i not in self.ech.N_I:
+            c_len += 1
+        if c_len == 0:  # pragma: no cover
+            msg = 'PrimalSageCone constraint "' + self.name + '" encountered an index '
+            msg += 'i=' + str(i) + '\n where the i-th AGE cone reduces to the nonnegative '
+            msg += 'orthant, but self.c[i]=' + str(self.c[i].value) + ' is negative. \n\n'
+            msg += 'This SAGE constraint is infeasible.'
+            raise RuntimeError(msg)
+        return c_len
+
     def _build_aligned_age_vectors(self):
         if self._m > 1:
             for i in self.ech.U_I:
@@ -195,7 +288,7 @@ class PrimalSageCone(SetMembership):
         aux_c_vars = aux_c_vars[nonconst_locs, :]
         main_c_var = self.c[nonconst_locs]
         A_vals, A_rows, A_cols, b = comp_aff.columns_sum_leq_vec(aux_c_vars, main_c_var)
-        conetype = '0' if _SUM_AGE_FORCE_EQUALITY_ else '+'
+        conetype = '0' if self.settings['sum_age_force_equality'] else '+'
         K = [Cone(conetype, b.size)]
         return A_vals, A_rows, A_cols, b, K
 
@@ -203,7 +296,7 @@ class PrimalSageCone(SetMembership):
         return self._variables
 
     def conic_form(self):
-        if len(self._nu_vars) == 0:
+        if len(self._nus) == 0:
             cd = self._trivial_conic_form()
         else:
             if self.X is None:
@@ -221,23 +314,25 @@ class PrimalSageCone(SetMembership):
 
     def _ordsage_conic_form(self):
         cone_data = []
+        nu_keys = self._nus.keys()
         for i in self.ech.U_I:
-            if i in self._nu_vars:
+            if i in nu_keys:
                 idx_set = self.ech.expcovers[i]
                 # relative entropy inequality constraint
-                x = self._nu_vars[i]
+                x = self._nus[i]
                 y = np.exp(1) * self.age_vectors[i][idx_set]  # This line consumes a large amount of runtime
                 z = -self.age_vectors[i][i]
                 epi = self._relent_epi_vars[i]
                 cd = sum_relent(x, y, z, epi)
                 cone_data.append(cd)
-                # linear equality constraints
-                mat = (self.alpha[idx_set, :] - self.alpha[i, :]).T
-                av, ar, ac = comp_aff.mat_times_vecvar(mat, self._nu_vars[i])
-                num_rows = mat.shape[0]
-                curr_b = np.zeros(num_rows, )
-                curr_k = [Cone('0', num_rows)]
-                cone_data.append((av, ar, ac, curr_b, curr_k))
+                if not self.settings['kernel_basis']:
+                    # linear equality constraints
+                    mat = (self.alpha[idx_set, :] - self.alpha[i, :]).T
+                    av, ar, ac = comp_aff.mat_times_vecvar(mat, self._nus[i])
+                    num_rows = mat.shape[0]
+                    curr_b = np.zeros(num_rows, )
+                    curr_k = [Cone('0', num_rows)]
+                    cone_data.append((av, ar, ac, curr_b, curr_k))
             else:
                 con = 0 <= self.age_vectors[i][i]
                 con.epigraph_checked = True
@@ -253,10 +348,10 @@ class PrimalSageCone(SetMembership):
             zero_block = np.zeros(shape=(self._m, self._lifted_n - self._n))
             lifted_alpha = np.hstack((lifted_alpha, zero_block))
         for i in self.ech.U_I:
-            if i in self._nu_vars:
+            if i in self._nus:
                 idx_set = self.ech.expcovers[i]
                 # relative entropy inequality constraint
-                x = self._nu_vars[i]
+                x = self._nus[i]
                 y = np.exp(1) * self.age_vectors[i][idx_set]  # takes weirdly long amount of time.
                 z = -self.age_vectors[i][i] + self._eta_vars[i] @ self.X.b
                 epi = self._relent_epi_vars[i]
@@ -265,7 +360,7 @@ class PrimalSageCone(SetMembership):
                 # linear equality constraints
                 mat1 = (lifted_alpha[idx_set, :] - lifted_alpha[i, :]).T
                 mat2 = -self.X.A.T
-                var1 = self._nu_vars[i]
+                var1 = self._nus[i]
                 var2 = self._eta_vars[i]
                 av, ar, ac = comp_aff.mat_times_vecvar_plus_mat_times_vecvar(mat1, var1, mat2, var2)
                 num_rows = mat1.shape[0]
@@ -298,60 +393,117 @@ class PrimalSageCone(SetMembership):
         return prob.value
 
     def violation(self, norm_ord=np.inf, rough=False):
+        """
+        Return a measure of violation for the constraint that ``self.c`` belongs to
+        :math:`C_{\\mathrm{SAGE}}(\\alpha, X)`.
+
+        Parameters
+        ----------
+        norm_ord : int
+            The value of ``ord`` passed to numpy ``norm`` functions, when reducing
+            vector-valued residuals into a scalar residual.
+
+        rough : bool
+            Setting ``rough=False`` computes violation by solving an optimization
+            problem. Setting ``rough=True`` computes violation by taking norms of
+            residuals of appropriate elementwise equations and inequalities involving
+            ``self.c`` and auxiliary variables.
+
+        Notes
+        -----
+        When ``rough=False``, the optimization-based violation is computed by projecting
+        the vector ``self.c`` onto a new copy of a primal SAGE constraint, and then returning
+        the L2-norm between ``self.c`` and that projection. This optimization step essentially
+        re-solves for all auxiliary variables used by this constraint.
+
+        """
         c = self.c.value
-        if self._m > 1:
-            if not rough:
-                dist = PrimalSageCone.project(c, self.alpha, self.X)
-                return dist
-            # compute violation for "AGE vectors sum to c"
-            #   Although, we can use the fact that the SAGE cone contains R^m_++.
-            #   and so only compute violation for "AGE vectors sum to <= c"
-            age_vectors = {i: v.value for i, v in self.age_vectors.items()}
-            sum_age_vectors = sum(age_vectors.values())
-            residual = c - sum_age_vectors  # want >= 0
-            residual[residual > 0] = 0
-            sum_to_c_viol = np.linalg.norm(residual, ord=norm_ord)
-            # compute violations for each AGE cone
-            alpha = self.alpha
-            if self._lifted_n > self._n:
-                # Then need to zero-pad alpha
-                zero_block = np.zeros(shape=(self._m, self._lifted_n - self._n))
-                alpha = np.hstack((alpha, zero_block))
-            age_viols = []
-            for i in self.ech.U_I:
-                if i in self._nu_vars:
-                    eta_vec = 0 if self.X is None else self._eta_vars[i].value
-                    eta_viol = 0 if self.X is None else DualProductCone.project(eta_vec, self.X.K)
-                    c_i = self.age_vectors[i].value
-                    x_i = self._nu_vars[i].value
-                    x_i[x_i < 0] = 0
-                    idx_set = self.ech.expcovers[i]
-                    y_i = np.exp(1) * c_i[idx_set]
-                    condsage_term = 0 if self.X is None else self.X.b @ eta_vec
-                    relent_res = np.sum(special_functions.rel_entr(x_i, y_i)) - c_i[i] + condsage_term  # <= 0
-                    relent_viol = 0 if relent_res < 0 else relent_res
-                    condsage_term = 0 if self.X is None else self.X.A.T @ eta_vec
-                    eq_res = (alpha[idx_set, :] - alpha[i, :]).T @ x_i - condsage_term  # == 0
-                    eq_res = eq_res.reshape((-1,))
-                    eq_viol = np.linalg.norm(eq_res, ord=norm_ord)
-                    total_viol = relent_viol + eq_viol + eta_viol
-                    age_viols.append(total_viol)
-                else:
-                    c_i = float(self._c_vars[i].value)
-                    relent_viol = 0 if c_i >= 0 else -c_i
-                    age_viols.append(relent_viol)
-            age_viols = np.array(age_viols)
-            # add the max "AGE violation" to the violation for "AGE vectors sum to c".
-            if np.any(age_viols == np.inf):
-                total_viol = sum_to_c_viol + np.sum(age_viols[age_viols < np.inf])
-                total_viol += PrimalSageCone.project(c, self.alpha, self.X)
-            else:
-                total_viol = sum_to_c_viol + np.max(age_viols)
-            return total_viol
-        else:
+        if self._m == 1:
             residual = c.reshape((-1,))  # >= 0
             residual[residual >= 0] = 0
             return np.linalg.norm(c, ord=norm_ord)
+        if not rough:
+            dist = PrimalSageCone.project(c, self.alpha, self.X)
+            return dist
+        # compute violation for "AGE vectors sum to c"
+        #   Although, we can use the fact that the SAGE cone contains R^m_++.
+        #   and so only compute violation for "AGE vectors sum to <= c"
+        age_vectors = {i: v.value for i, v in self.age_vectors.items()}
+        sum_age_vectors = sum(age_vectors.values())
+        residual = c - sum_age_vectors  # want >= 0
+        residual[residual > 0] = 0
+        sum_to_c_viol = np.linalg.norm(residual, ord=norm_ord)
+        # compute violations for each AGE cone
+        alpha = self.alpha
+        if self._lifted_n > self._n:
+            # Then need to zero-pad alpha
+            zero_block = np.zeros(shape=(self._m, self._lifted_n - self._n))
+            alpha = np.hstack((alpha, zero_block))
+        age_viols = []
+        nu_keys = self._nus.keys()
+        for i in self.ech.U_I:
+            if i in nu_keys:
+                c_i = self.age_vectors[i].value
+                x_i = self._nus[i].value
+                x_i[x_i < 0] = 0
+                idx_set = self.ech.expcovers[i]
+                sf_part = self.sigma_x((alpha[i, :] - alpha[idx_set, :]).T @ x_i)
+                y_i = np.exp(1) * c_i[idx_set]
+                relent_res = np.sum(special_functions.rel_entr(x_i, y_i)) - c_i[i] + sf_part  # <= 0
+                relent_viol = 0 if relent_res < 0 else relent_res
+                age_viols.append(relent_viol)
+            else:
+                c_i = float(self._c_vars[i].value)
+                relent_viol = 0 if c_i >= 0 else -c_i
+                age_viols.append(relent_viol)
+        age_viols = np.array(age_viols)
+        # add the max "AGE violation" to the violation for "AGE vectors sum to c".
+        if np.any(age_viols == np.inf):
+            total_viol = sum_to_c_viol + np.sum(age_viols[age_viols < np.inf])
+            total_viol += PrimalSageCone.project(c, self.alpha, self.X)
+        else:
+            total_viol = sum_to_c_viol + np.max(age_viols)
+        return total_viol
+
+    def sigma_x(self, y, tol=1e-8):
+        """
+        If :math:`X = \\mathbb{R}^n`, then return :math:`\\infty` when :math:`\|y\|_2 > \\texttt{tol}`
+        and :math:`0` otherwise. If :math:`X` is a proper subset of :math:`\\mathbb{R}^n`, then
+        return the value of the support function of :math:`X`, evaluated at :math:`y`:
+
+        .. math::
+
+            \\sigma_X(y) \\doteq \\max\\{ y^\\intercal x \\,:\\, x \\in X \\}.
+
+        See also, the attribute ``PrimalSageCone.age_witnesses``.
+        """
+        if isinstance(y, Expression):
+            y = y.value
+        if self.X is None:
+            res = np.linalg.norm(y)
+            if res > tol:
+                return np.inf
+            return 0.0
+        val = self.X.suppfunc(y)
+        return val
+
+    @property
+    def age_witnesses(self):
+        if self._age_witnesses is None:
+            self._age_witnesses = dict()
+            if self._m > 1:
+                for i in self.ech.U_I:
+                    wi_expr = Expression(np.zeros(self._m,))
+                    num_cover = self.ech.expcover_counts[i]
+                    if num_cover > 0:
+                        x_i = self._nus[i]
+                        wi_expr[self.ech.expcovers[i]] = pos_operator(x_i, eval_only=True)
+                        wi_expr[i] = -aff.sum(wi_expr[self.ech.expcovers[i]])
+                    self._age_witnesses[i] = wi_expr
+            else:
+                self.age_vectors[0] = self.c
+                self._age_witnesses[0] = Expression([0])
+        return self._age_witnesses
 
 
 class DualSageCone(SetMembership):
@@ -417,18 +569,21 @@ class DualSageCone(SetMembership):
         solution recovery algorithm takes these variables, and considers points ``x`` of
         the form ``x = mu_vars[i].value / self.v[i].value``.
 
-    ech : ExpCoverHelper
-
-        A simple wrapper around the constructor argument ``covers``. Manages validation of ``covers``
-        when provided, and manages construction of ``covers`` when a user does not provide it.
-
     name : str
 
         Uniquely identifies this Constraint in the model where it appears. Serves as a suffix
         for the name of any auxiliary Variable created when compiling to the coniclifts standard.
+
+    ech : ExpCoverHelper
+
+        A simple wrapper around the constructor argument ``covers``. Validates ``covers``
+        when provided, and constructs ``covers`` when a user does not provide it.
     """
 
     def __init__(self, v, alpha, X, name, **kwargs):
+        self.settings = SETTINGS.copy()
+        if 'settings' in kwargs:
+            self.settings.update(kwargs['settings'])
         covers = kwargs['covers'] if 'covers' in kwargs else None
         c = kwargs['c'] if 'c' in kwargs else None
         self.alpha = alpha
@@ -439,10 +594,10 @@ class DualSageCone(SetMembership):
         if X is not None:
             check_cones(X.K)
             self._lifted_n = X.A.shape[1]
-            self.ech = ExpCoverHelper(self.alpha, self.c, (X.A, X.b, X.K), covers)
+            self.ech = ExpCoverHelper(self.alpha, self.c, (X.A, X.b, X.K), covers, self.settings)
         else:
             self._lifted_n = self._n
-            self.ech = ExpCoverHelper(self.alpha, self.c, None, covers)
+            self.ech = ExpCoverHelper(self.alpha, self.c, None, covers, self.settings)
         self.X = X
         self.mu_vars = dict()
         self.name = name
@@ -460,7 +615,7 @@ class DualSageCone(SetMembership):
                 self._variables.append(self._lifted_mu_vars[i])
                 self.mu_vars[i] = self._lifted_mu_vars[i][:self._n]
                 num_cover = self.ech.expcover_counts[i]
-                if num_cover > 0:
+                if num_cover > 0 and not self.settings['compact_dual']:
                     var_name = '_relent_epi_[' + str(i) + ']_{' + self.name + '}'
                     epi = Variable(shape=(num_cover,), name=var_name)
                     self._relent_epi_vars[i] = epi
@@ -483,19 +638,24 @@ class DualSageCone(SetMembership):
                 num_cover = self.ech.expcover_counts[i]
                 if num_cover == 0:
                     continue
-                # relative entropy constraints
                 expr = np.tile(self.v[i], num_cover).view(Expression)
-                epi = self._relent_epi_vars[i]
-                cd = elementwise_relent(expr, self.v[idx_set], epi)
-                cone_data.append(cd)
-                # Linear inequalities
-                mat = self.alpha[idx_set, :] - self.alpha[i, :]
+                mat = self.alpha[i, :] - self.alpha[idx_set, :]
                 vecvar = self._lifted_mu_vars[i][:self._n]
-                av, ar, ac = comp_aff.mat_times_vecvar_minus_vecvar(-mat, vecvar, epi)
-                num_rows = mat.shape[0]
-                curr_b = np.zeros(num_rows)
-                curr_k = [Cone('+', num_rows)]
-                cone_data.append((av, ar, ac, curr_b, curr_k))
+                if self.settings['compact_dual']:
+                    epi = mat @ vecvar
+                    cd = elementwise_relent(expr, self.v[idx_set], epi)
+                    cone_data.append(cd)
+                else:
+                    # relative entropy constraints
+                    epi = self._relent_epi_vars[i]
+                    cd = elementwise_relent(expr, self.v[idx_set], epi)
+                    cone_data.append(cd)
+                    # Linear inequalities
+                    av, ar, ac = comp_aff.mat_times_vecvar_minus_vecvar(mat, vecvar, epi)
+                    num_rows = mat.shape[0]
+                    curr_b = np.zeros(num_rows)
+                    curr_k = [Cone('+', num_rows)]
+                    cone_data.append((av, ar, ac, curr_b, curr_k))
                 # membership in cone induced by self.AbK
                 if self.X is not None:
                     A, b, K = self.X.A, self.X.b, self.X.K
@@ -503,8 +663,7 @@ class DualSageCone(SetMembership):
                     singlevar = self.v[i]
                     av, ar, ac = comp_aff.mat_times_vecvar_plus_vec_times_singlevar(A, vecvar, b, singlevar)
                     curr_b = np.zeros(b.size, )
-                    curr_k = [Cone(co.type, co.len) for co in K]
-                    cone_data.append((av, ar, ac, curr_b, curr_k))
+                    cone_data.append((av, ar, ac, curr_b, K))
             return cone_data
         else:
             con = self.v >= 0
@@ -514,6 +673,29 @@ class DualSageCone(SetMembership):
             return cone_data
 
     def violation(self, norm_ord=np.inf, rough=False):
+        """
+        Return a measure of violation for the constraint that ``self.v`` belongs to
+        :math:`C_{\\mathrm{SAGE}}(\\alpha, X)^{\\dagger}`.
+
+        Parameters
+        ----------
+        norm_ord : int
+            The value of ``ord`` passed to numpy ``norm`` functions, when reducing
+            vector-valued residuals into a scalar residual.
+
+        rough : bool
+            Setting ``rough=False`` computes violation by solving an optimization
+            problem. Setting ``rough=True`` computes violation by taking norms of
+            residuals of appropriate elementwise equations and inequalities involving
+            ``self.v`` and auxiliary varibles.
+
+        Notes
+        -----
+        When ``rough=False``, the optimization-based violation is computed by projecting
+        the vector ``self.v`` onto a new copy of a dual SAGE constraint, and then returning
+        the L2-norm between ``self.v`` and that projection. This optimization step essentially
+        re-solves for all auxiliary variables used by this constraint.
+        """
         v = self.v.value
         viols = []
         for i in self.ech.U_I:
@@ -542,7 +724,7 @@ class DualSageCone(SetMembership):
                         cons.append(con)
                     prob = Problem(CL_MIN, Expression([0]), cons)
                     status, value = prob.solve(verbose=False)
-                    if status == CL_SOLVED and abs(value) < 1e-7:
+                    if status in {CL_SOLVED, CL_INACCURATE} and abs(value) < 1e-7:
                         curr_viol = 0
                 viols.append(curr_viol)
             else:
@@ -553,7 +735,10 @@ class DualSageCone(SetMembership):
 
 class ExpCoverHelper(object):
 
-    def __init__(self, alpha, c, AbK, expcovers=None):
+    def __init__(self, alpha, c, AbK, expcovers=None, settings=None):
+        self.settings = SETTINGS.copy()
+        if settings is not None:
+            self.settings.update(settings)
         if c is not None and not isinstance(c, Expression):
             raise RuntimeError()
         self.m = alpha.shape[0]
@@ -613,7 +798,7 @@ class ExpCoverHelper(object):
             cov[self.N_I] = False
             cov[i] = False
             expcovers[i] = cov
-        if self.AbK is None or _AGGRESSIVE_REDUCTION_:
+        if self.AbK is None or self.settings['heuristic_reduction']:
             row_sums = np.sum(self.alpha, 1)
             if np.all(self.alpha >= 0) and np.min(row_sums) == 0:
                 # Then apply the reduction.
@@ -678,7 +863,7 @@ class ExpCoverHelper(object):
             for i in self.U_I:
                 if np.count_nonzero(expcovers[i]) == 1:
                     expcovers[i][:] = False
-        if _ELIMINATE_TRIVIAL_AGE_CONES_:
+        if self.settings['presolve_trivial_age_cones']:
             if self.AbK is None:
                 for i in self.U_I:
                     if np.any(expcovers[i]):
@@ -687,8 +872,9 @@ class ExpCoverHelper(object):
                         objective = Expression([0])
                         cons = [mat @ x <= -1]
                         prob = Problem(CL_MIN, objective, cons)
-                        prob.solve(verbose=False, solver=_REDUCTION_SOLVER_)
-                        if prob.status == CL_SOLVED and abs(prob.value) < 1e-7:
+                        prob.solve(verbose=False,
+                                   solver=self.settings['reduction_solver'])
+                        if prob.status in {CL_SOLVED, CL_INACCURATE} and abs(prob.value) < 1e-7:
                             expcovers[i][:] = False
             else:
                 for i in self.U_I:
@@ -700,8 +886,9 @@ class ExpCoverHelper(object):
                         A, b, K = self.AbK
                         cons = [mat @ x <= t, PrimalProductCone(A @ x + b, K)]
                         prob = Problem(CL_MIN, objective, cons)
-                        prob.solve(verbose=False, solver=_REDUCTION_SOLVER_)
-                        if prob.status == CL_SOLVED and prob.value < -100:
+                        prob.solve(verbose=False,
+                                   solver=self.settings['reduction_solver'])
+                        if prob.status == {CL_SOLVED, CL_INACCURATE} and prob.value < -100:
                             expcovers[i][:] = False
         return expcovers
 
