@@ -18,6 +18,7 @@ from sageopt import coniclifts as cl
 from sageopt.symbolic.polynomials import Polynomial, PolyDomain
 from sageopt.symbolic.signomials import Signomial
 from sageopt.relaxations import sage_sigs
+from sageopt.relaxations.sage_sigs import _check_kwargs
 from sageopt.relaxations import constraint_generators as con_gen
 from sageopt.relaxations import symbolic_correspondences as sym_corr
 
@@ -94,6 +95,7 @@ def poly_relaxation(f, X=None, form='dual', **kwargs):
     by this function. If you want to recover solutions, call ``poly_constrained_relaxation``
     with empty lists ``gts=[]`` and  ``eqs=[]``.
     """
+    _check_kwargs(kwargs, allowed={'poly_ell', 'sigrep_ell'})
     poly_ell = kwargs['poly_ell'] if 'poly_ell' in kwargs else 0
     sigrep_ell = kwargs['sigrep_ell'] if 'sigrep_ell' in kwargs else 0
     form = form.lower()
@@ -121,7 +123,7 @@ def poly_dual(f, poly_ell=0, sigrep_ell=0, X=None):
         constraints = relative_dual_sage_poly_cone(lagrangian, v, con_base_name, log_AbK=X)
         a = sym_corr.relative_coeff_vector(modulator, lagrangian.alpha)
         constraints.append(a.T @ v == 1)
-        f_mod = Polynomial(f.alpha_c) * modulator
+        f_mod = Polynomial(f.alpha, f.c) * modulator
         obj_vec = sym_corr.relative_coeff_vector(f_mod, lagrangian.alpha)
         obj = obj_vec.T @ v
         prob = cl.Problem(cl.MIN, obj, constraints)
@@ -200,9 +202,9 @@ def sage_multiplier_search(f, level=1, X=None):
 
     Notes
     -----
-    This function provides an alternative to moving up the SAGE hierarchy, for the goal of certifying
-    nonnegativity of a polynomial ``f`` over some set ``X`` where ``|X|`` is log-convex.
-    In general, the approach is to introduce a polynomial
+    This function provides an alternative to moving up the reference SAGE hierarchy, for the
+    goal of certifying nonnegativity of a polynomial ``f`` over some set ``X`` where ``|X|``
+    is log-convex. In general, the approach is to introduce a polynomial
 
         ``mult = Polynomial(alpha_hat, c_tilde)``
 
@@ -279,6 +281,7 @@ def poly_constrained_relaxation(f, gts, eqs, X=None, form='dual', **kwargs):
         Controls the strength of the SAGE proof system, as applied to the Lagrangian. Defaults to
         ``ell=0``, which means the primal Lagrangian must be an X-SAGE polynomial.
     """
+    _check_kwargs(kwargs, allowed={'p', 'q', 'ell'})
     form = form.lower()
     p = kwargs['p'] if 'p' in kwargs else 0
     q = kwargs['q'] if 'q' in kwargs else 1
@@ -337,12 +340,12 @@ def poly_constrained_dual(f, gts, eqs, p=0, q=1, ell=0, X=None):
     lagrangian, ineq_lag_mults, eq_lag_mults, _ = make_poly_lagrangian(f, gts, eqs, p=p, q=q)
     metadata = {'lagrangian': lagrangian, 'f': f, 'gts': gts, 'eqs': eqs, 'X': X}
     if ell > 0:
-        alpha_E_1 = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
+        alpha_E_1 = hierarchy_e_k([f, f.upcast_to_polynomial(1)] + gts + eqs, k=1)
         modulator = Polynomial(2 * alpha_E_1, np.ones(alpha_E_1.shape[0])) ** ell
         lagrangian = lagrangian * modulator
         f = f * modulator
     else:
-        modulator = Polynomial({(0,) * f.n: 1})
+        modulator = f.upcast_to_polynomial(1)
     metadata['modulator'] = modulator
     # In primal form, the Lagrangian is constrained to be a SAGE polynomial.
     # Introduce a dual variable "v" for this constraint.
@@ -457,7 +460,7 @@ def make_poly_lagrangian(f, gts, eqs, p, q):
     folded_gt = con_gen.up_to_q_fold_cons(gts, q)
     gamma = cl.Variable(name='gamma')
     L = f - gamma
-    alpha_E_p = hierarchy_e_k([f] + list(gts) + list(eqs), k=p)
+    alpha_E_p = hierarchy_e_k([f, f.upcast_to_polynomial(1)] + gts + eqs, k=p)
     alpha_multiplier = np.vstack([2 * alpha_E_p, alpha_E_p])
     alpha_multiplier = np.unique(alpha_multiplier, axis=0)
     ineq_dual_polys = []
@@ -501,12 +504,12 @@ def infer_domain(f, gts, eqs, check_feas=True):
     """
     # GP-representable inequality constraints (recast as "Signomial >= 0")
     gp_gts = con_gen.valid_gp_representable_poly_inequalities(gts)
-    gp_gts_sigreps = [Signomial(g.alpha_c) for g in gp_gts]
+    gp_gts_sigreps = [Signomial(g.alpha, g.c) for g in gp_gts]
     gp_gts_sigreps = con_gen.valid_posynomial_inequalities(gp_gts_sigreps)
     #   ^ That second call is to convexify the signomials.
     # GP-representable equality constraints (recast as "Signomial == 0")
     gp_eqs = con_gen.valid_gp_representable_poly_eqs(eqs)
-    gp_eqs_sigreps = [Signomial(g.alpha_c) for g in gp_eqs]
+    gp_eqs_sigreps = [Signomial(g.alpha, g.c) for g in gp_eqs]
     gp_eqs_sigreps = con_gen.valid_monomial_equations(gp_eqs_sigreps)
     #  ^ That second call is to make sure the nonconstant term has
     #    a particular sign (specifically, a negative sign).
@@ -521,9 +524,11 @@ def infer_domain(f, gts, eqs, check_feas=True):
 
 
 def hierarchy_e_k(polys, k):
-    alpha_tups = sum([list(s.alpha_c.keys()) for s in polys], [])
-    alpha_tups = set(alpha_tups)
-    s = Polynomial(dict([(a, 1.0) for a in alpha_tups]))
+    alphas = [s.alpha for s in polys]
+    alpha = np.vstack(alphas)
+    alpha = np.unique(alpha, axis=0)
+    c = np.ones(shape=(alpha.shape[0],))
+    s = Polynomial(alpha, c)
     s = s ** k
     return s.alpha
 
