@@ -38,9 +38,11 @@ def sig_relaxation(f, X=None, form='dual', **kwargs):
 
         f_X^{\\star} \doteq \min\{ f(x) \,:\, x \\in X \}
 
-    where X = :math:`R^{\\texttt{f.n}}` by default.
+    where X = :math:`\\mathbb{R}^{\\texttt{f.n}}` by default.
 
-    If ``form='dual'``, we can also attempt to recover solutions to the above problem.
+    When ``form='dual'``, a solution to this convex relaxation can be used to
+    recover optimal solutions to the problem above. Refer to the Notes for keyword
+    arguments accepted by this function.
 
     Parameters
     ----------
@@ -63,12 +65,13 @@ def sig_relaxation(f, X=None, form='dual', **kwargs):
     This function also accepts the following keyword arguments:
 
     ell : int
-        The level of the SAGE hierarchy. Must be nonnegative.
+        The level of the reference SAGE hierarchy. Must be nonnegative.
 
     mod_supp : NumPy ndarray
         Only used when ``ell > 0``. If ``mod_supp`` is not None, then the rows of this
-        array define the exponents of a positive definite modulating Signomial ``t`` in the SAGE hierarchy.
+        array define the exponents of a positive definite modulating Signomial ``t`` in the reference SAGE hierarchy.
     """
+    _check_kwargs(kwargs, allowed={'ell', 'mod_supp'})
     ell = kwargs['ell'] if 'ell' in kwargs else 0
     mod_supp = kwargs['mod_supp'] if 'mod_supp' in kwargs else None
     if form.lower()[0] == 'd':
@@ -81,12 +84,12 @@ def sig_relaxation(f, X=None, form='dual', **kwargs):
 
 
 def sig_dual(f, ell=0, X=None, modulator_support=None):
-    f.remove_terms_with_zero_as_coefficient()
+    f = f.without_zeros()
     # Signomial definitions (for the objective).
-    if modulator_support is None:
-        modulator_support = f.alpha
-    t_mul = Signomial(modulator_support, np.ones(modulator_support.shape[0])) ** ell
     lagrangian = f - cl.Variable(name='gamma')
+    if modulator_support is None:
+        modulator_support = lagrangian.alpha
+    t_mul = Signomial(modulator_support, np.ones(modulator_support.shape[0])) ** ell
     metadata = {'f': f, 'lagrangian': lagrangian, 'modulator': t_mul, 'X': X}
     lagrangian = lagrangian * t_mul
     f_mod = f * t_mul
@@ -109,13 +112,13 @@ def sig_dual(f, ell=0, X=None, modulator_support=None):
 
 
 def sig_primal(f, ell=0, X=None, modulator_support=None):
-    f.remove_terms_with_zero_as_coefficient()
-    if modulator_support is None:
-        modulator_support = f.alpha
-    t = Signomial(modulator_support, np.ones(modulator_support.shape[0]))
+    f = f.without_zeros()
     gamma = cl.Variable(name='gamma')
-    s_mod = (f - gamma) * (t ** ell)
-    s_mod.remove_terms_with_zero_as_coefficient()
+    lagrangian = f - gamma
+    if modulator_support is None:
+        modulator_support = lagrangian.alpha
+    t = Signomial(modulator_support, np.ones(modulator_support.shape[0]))
+    s_mod = lagrangian * (t ** ell)
     con = primal_sage_cone(s_mod, name=str(s_mod), X=X)
     constraints = [con]
     obj = gamma.as_expr()
@@ -145,7 +148,7 @@ def sage_feasibility(f, X=None, additional_cons=None):
         A coniclifts maximization Problem. If ``f`` admits an X-SAGE decomposition, then we should have
         ``prob.value > -np.inf``, once ``prob.solve()`` has been called.
     """
-    f.remove_terms_with_zero_as_coefficient()
+    f = f.without_zeros()
     con = primal_sage_cone(f, name=str(f), X=X)
     constraints = [con]
     if additional_cons is not None:
@@ -176,7 +179,7 @@ def sage_multiplier_search(f, level=1, X=None):
 
     Notes
     -----
-    This function provides an alternative to moving up the SAGE hierarchy, for the goal of certifying
+    This function provides an alternative to moving up the reference SAGE hierarchy, for the goal of certifying
     nonnegativity of a signomial ``f`` over some convex set ``X``.  In general, the approach is to introduce
     a signomial
 
@@ -186,9 +189,9 @@ def sage_multiplier_search(f, level=1, X=None):
     is a coniclifts Variable defining a nonzero X-SAGE function. Then we check if ``f_mod = f * mult``
     is X-SAGE for any choice of ``c_tilde``.
     """
-    f.remove_terms_with_zero_as_coefficient()
+    f = f.without_zeros()
     constraints = []
-    mult_alpha = hierarchy_e_k([f], k=level)
+    mult_alpha = hierarchy_e_k([f, f.upcast_to_signomial(1)], k=level)
     c_tilde = cl.Variable(mult_alpha.shape[0], name='c_tilde')
     mult = Signomial(mult_alpha, c_tilde)
     constraints.append(cl.sum(c_tilde) >= 1)
@@ -251,6 +254,7 @@ def sig_constrained_relaxation(f, gts, eqs, X=None, form='dual', **kwargs):
         Controls the strength of the SAGE proof system, as applied to the Lagrangian. Defaults to
         ``ell=0``, which means the primal Lagrangian must be an X-SAGE signomial.
     """
+    _check_kwargs(kwargs, allowed={'p', 'q', 'ell'})
     p = kwargs['p'] if 'p' in kwargs else 0
     q = kwargs['q'] if 'q' in kwargs else 1
     ell = kwargs['ell'] if 'ell' in kwargs else 0
@@ -278,11 +282,11 @@ def sig_constrained_primal(f, gts, eqs, p=0, q=1, ell=0, X=None):
     lagrangian, ineq_lag_mults, _, gamma = make_sig_lagrangian(f, gts, eqs, p=p, q=q)
     metadata = {'lagrangian': lagrangian, 'X': X}
     if ell > 0:
-        alpha_E_1 = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
+        alpha_E_1 = hierarchy_e_k([f, f.upcast_to_signomial(1)] + gts + eqs, k=1)
         modulator = Signomial(alpha_E_1, np.ones(alpha_E_1.shape[0])) ** ell
         lagrangian = lagrangian * modulator
     else:
-        modulator = Signomial({(0,) * f.n: 1})
+        modulator = f.upcast_to_signomial(1)
     metadata['modulator'] = modulator
     # The Lagrangian (after possible multiplication, as above) must be a SAGE signomial.
     con = primal_sage_cone(lagrangian, name='Lagrangian is SAGE', X=X)
@@ -314,12 +318,12 @@ def sig_constrained_dual(f, gts, eqs, p=0, q=1, ell=0, X=None):
     lagrangian, ineq_lag_mults, eq_lag_mults, _ = make_sig_lagrangian(f, gts, eqs, p=p, q=q)
     metadata = {'lagrangian': lagrangian, 'f': f, 'gts': gts, 'eqs': eqs, 'level': (p, q, ell), 'X': X}
     if ell > 0:
-        alpha_E_1 = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
+        alpha_E_1 = hierarchy_e_k([f, f.upcast_to_signomial(1)] + list(gts) + list(eqs), k=1)
         modulator = Signomial(alpha_E_1, np.ones(alpha_E_1.shape[0])) ** ell
         lagrangian = lagrangian * modulator
         f = f * modulator
     else:
-        modulator = Signomial({(0,) * f.n: 1})
+        modulator = f.upcast_to_signomial(1)
     metadata['modulator'] = modulator
     # In primal form, the Lagrangian is constrained to be a SAGE signomial.
     # Introduce a dual variable "v" for this constraint.
@@ -427,20 +431,22 @@ def make_sig_lagrangian(f, gts, eqs, p, q):
     folded_gt = con_gen.up_to_q_fold_cons(gts, q)
     gamma = cl.Variable(name='gamma')
     L = f - gamma
-    alpha_E_p = hierarchy_e_k([f] + list(gts) + list(eqs), k=p)
+    alpha_E_p = hierarchy_e_k([L] + list(gts) + list(eqs), k=p)
     ineq_dual_sigs = []
+    summands = [L]
     for g in folded_gt:
         s_g_coeff = cl.Variable(name='s_' + str(g), shape=(alpha_E_p.shape[0],))
         s_g = Signomial(alpha_E_p, s_g_coeff)
-        L -= s_g * g
+        summands.append(-g * s_g)
         ineq_dual_sigs.append((s_g, g))
     eq_dual_sigs = []
     folded_eq = con_gen.up_to_q_fold_cons(eqs, q)
     for g in folded_eq:
         z_g_coeff = cl.Variable(name='z_' + str(g), shape=(alpha_E_p.shape[0],))
         z_g = Signomial(alpha_E_p, z_g_coeff)
-        L -= z_g * g
+        summands.append(-g * z_g)
         eq_dual_sigs.append((z_g, g))
+    L = Signomial.sum(summands)
     return L, ineq_dual_sigs, eq_dual_sigs, gamma
 
 
@@ -477,8 +483,16 @@ def infer_domain(f, gts, eqs, check_feas=True):
 
 
 def hierarchy_e_k(sigs, k):
-    alpha_tups = sum([list(s.alpha_c.keys()) for s in sigs], [])
-    alpha_tups = set(alpha_tups)
-    s = Signomial(dict([(a, 1.0) for a in alpha_tups]))
+    alphas = [s.alpha for s in sigs]
+    alpha = np.vstack(alphas)
+    alpha = np.unique(alpha, axis=0)
+    c = np.ones(shape=(alpha.shape[0],))
+    s = Signomial(alpha, c)
     s = s ** k
     return s.alpha
+
+
+def _check_kwargs(kwargs, allowed):
+    for kw_key in kwargs:
+        if kw_key not in allowed:
+            raise ValueError('Keyword argument "' + kw_key + '" not recognized.')
